//
//  Wed May 04 15:25:02 2005 -- Scott R. Turner
//
//  Welcome to the source code for Platypus.  Please wipe your feet
//  before entering.  We hope you'll enjoy your stay.
//
//  Copyright (c) 2005, Scott R. Turner
//  Released under the GPL license
//  http://www.gnu.org/copyleft/gpl.html
//

//
//  Global variables
//
//  Mon May 09 13:13:16 2005 -- Scott R. Turner
//
//  Global variables are causing a problem because apparently Firefox
//  doesn't create a separate extension for window, but just runs a
//  single instance.  Damn the Firefox developers for their sensible
//  design decisions!  Anyway, that means we have to get rid of all
//  the lovely global variables, and make sure we somehow access
//  everything through "window".

//
//  Debugging and Utility Functions.  Not that I need them.
//

//
//  Dump
//  Sun May 01 22:44:52 2005 -- Scott R. Turner
//  Drop a message on the console.
//
function Dump(aMessage) {
    try {
	var consoleService =
	    Components.classes["@mozilla.org/consoleservice;1"].
	    getService(Components.interfaces.nsIConsoleService);
	var pref = Components.classes["@mozilla.org/preferences-service;1"].
	    getService(Components.interfaces.nsIPrefService).
	    getBranch("platypus.");
	if (pref.getPrefType("debug") == pref.PREF_BOOL &&
	    pref.getBoolPref("debug") == true)
	    consoleService.logStringMessage("Platypus: " + aMessage);
    } catch (e) { };
}

//
//  Warn
//  Fri May 13 09:34:54 2005 -- Scott R. Turner
//  Same thing, but regardless of debug status.
//
function Warning(aMessage) {
    try {
	var consoleService =
	    Components.classes["@mozilla.org/consoleservice;1"].
	    getService(Components.interfaces.nsIConsoleService);
	consoleService.logStringMessage("Platypus: " + aMessage);
    } catch (e) { };
}

//
//  Thu Apr 28 08:16:52 2005 -- Scott Turner
// 
//  My version of a generic tree-walking function.
//
function walk_down(node, func) {
  if (node.nodeType == 1) {
    if (node.tagName != "IMG") func(node);
    if (!(node.childNodes.length == 0 && 
	  leafElems[node.nodeName]))
      for (var i=0; i<node.childNodes.length; i++)
	walk_down(node.childNodes.item(i),func);
  }
}

//
//  Focus & Highlighting
//
//  Platypus keeps track of the element on the page the user is
//  manipulating; this is the focus.  The focus is kept highlighted
//  to give the user feedback.  At the moment we assume only one
//  element can be in focus; this might change in the future.
//

//
//  focus_on
//  Sat Apr 30 23:55:52 2005 -- Scott R. Turner
//
//  If the focus has really changed, we clear the highlight
//  on the current focus, put a highlight on the new element, and
//  then make the new element the focus.
//
//  Tue May 03 12:52:45 2005 -- Scott R. Turner
//
//  Don't do highlighting, etc., if we're replaying a script.
//
function focus_on(doc, elem) {
  Dump("In focus_on, interactive = "+
       doc.interactive+", elem = "+elem+
       " and focus = "+doc.focus);
  if (doc.interactive &&
      elem != doc.focus) {
    Dump(" ...calling clear().");
    platypus_clear(doc.focus);
    Dump(" ...calling highlight().");
    platypus_highlight(elem);
    doc.focus = elem;
  };
};

function clear_focus(doc) {
  platypus_clear(doc.focus);
  doc.focus = null;
};
  
//
//  Navigation Commands
//
//  Navigation is done primarily through the mouse, but I also provide
//  some functions for the arrow keys.  They don't always do what you
//  suspect -- I think that adds an air of mystery and delight to the
//  interface. 
//
function navigate_wider (doc, node) {
  var p = find_named_parent(node);
  if (doc.focus && p) {
    focus_on(doc, p);
  };
};

function navigate_narrower (doc, node) {
  var c = find_named_child(node);
  if (doc.focus && c) {
    focus_on(doc, c);
  };
};

function find_named_parent(node) {
  var parent = node.parentNode;
  if (parent) {
    if (parent.tagName &&
	(!parent.style || parent.style.display != "none")) return parent;
    return find_named_parent(parent);
  };
  return null;
};
  
function find_named_child(node) {
  if (node.childNodes && node.childNodes.length > 0) {
    for(var i=0;i<node.childNodes.length;i++) 
      if (node.childNodes[i].tagName &&
	  (!node.childNodes[i].style ||
	   node.childNodes[i].style.display != "none"))
	return node.childNodes[i];
  };
  return null;
};

//
//  Mon May 02 14:23:16 2005 -- Scott R. Turner
//
//  My new take on navigating forward and backward is to find the next
//  (or previous) node of the same type.  
//
//  Tue May 03 14:31:44 2005 -- Scott R. Turner
//
//  If you want to see something real interesting, try globally replacing
//  window._content.document in this file with window.document and
//  then repackaging and installing the extension!  Navigate around
//  the whole browser and have fun!
//
//  Fri May 06 10:25:10 2005 -- Scott R. Turner
//
//  Another re-re-vision of this function.  I've decided that it's
//  better for these functions to navigate at a particular tree level.
//
function navigate_next_sibling (doc, node) {
  var next_node = node.nextSibling;

  if (next_node == null) {
    return;
  };

  if (!next_node.tagName ||
      !next_node.style ||
      next_node.style.display == "none")
    return(navigate_next_sibling(doc, next_node));

  focus_on(doc, next_node);

};

//
//  Mon May 02 22:11:04 2005 -- Scott R. Turner
//
//  A little bit later on you're going to be reading about how stupid
//  I was when I originally wrote these functions.  I know the
//  temptation will be to laugh long and hearty at my foolishness, but
//  I beg you to have mercy.  The process of programming is like
//  taking a stinking heap of garbage and through relentness revision
//  and correction turning it into a Porsche Boxster.  So the fact
//  that I made some rather foolish errors in my original version of
//  these functions is proof positive that I'm a good programmer. In
//  fact, I purposely include at least a hundred errors in my first
//  draft of any program, just to give myself good material for
//  eventual success.
//
function navigate_previous_sibling (doc, node) {
  var next_node = node.previousSibling;
  
  if (next_node == null) return;

  if (!next_node.tagName ||
      !next_node.style ||
      next_node.style.display == "none")
    return(navigate_previous_sibling(doc, next_node));

  focus_on(doc, next_node);

};

//
//  User Commands
//
//  This section contains the functions that implement user commands.
//

//
//  Mon May 02 16:09:01 2005 -- Scott R. Turner
//
//  Erase by sticking in an empty box of the same dimensions and
//  then hiding the original element.  We have to use getComputedStyle
//  to get the actual displayed dimensions, as far as I can tell.
//  (This, by the way, turns out to be wrong.)
//
//  Mon May 02 21:08:46 2005 -- Scott R. Turner
//
//  You know, as I'm sitting here working on this code, I keep
//  reloading the extension, and every time the browser says "You
//  should only install software from sources you trust."  Well, I
//  sure as hell don't trust myself to write solid software -- the
//  reason I'm seeing that message so often is that I'm generating
//  more bugs than code.  So it's a real dilemma.
//
function erase(doc, node) {
  //
  //  Get height and width.  It turns out, strangely enough, that
  //  there's no compliant way to get the actual rendered height and
  //  width of an element.  offsetHeight and offsetWidth are IE
  //  standards that, fortunately for us, Firefox supports.
  //
  var offset_height = node.offsetHeight;
  var offset_width = node.offsetWidth;
  //
  //  Now create a DIV of the same size
  //
  var replacement_div = doc.createElement ("DIV");
  replacement_div.setAttribute('style',
			       "height: "+offset_height+"; width: "+offset_width+";");
  //
  //  Now stick him in right before (or after, doesn't matter) of
  //  the current element.
  //
  node.parentNode.insertBefore(replacement_div, node);
  //
  //  Now set display on the original node to none, making it vanish.
  //
  node.style.display = "none";
  //
  //  And switch the focus to the replacement.
  //
  focus_on(doc, replacement_div);
};

//
//  Mon May 02 16:03:07 2005 -- Scott R. Turner
//
//  Remove focused element from the page.
//
function remove(doc, node) {
  Dump("Called remove with: "+doc+" and "+node);
  //
  //  If no node or doc, maybe something has happened in the user
  //  script, but we have to bail.
  //
  if (doc == null || node == null) return;
  //
  //  Can't do anything if this node has no parent.  Sad little orphan
  //  node. 
  //
  if (!node.parentNode) return;
  //
  //  First move the focus up to the parent.
  //
  navigate_wider(doc, node);
  //
  //  Now set display=none for this element.  This is probably
  //  safer than actually deleting the element, since there's some
  //  (small) chance that there's a script or something buried in
  //  here that we'll need to keep the page working.
  //
  node.style.display = "none";
  doc.last_removed_node = node;
};

//
//  paste
//  Fri May 06 10:07:06 2005 -- Scott R. Turner
//
//  Paste in the last removed node.  This is actually a little
//  deceptive, since we didn't actually remove the node but only set
//  its display to "none".  To paste it in, what we'll actually do is
//  duplicate it and insert the duplicate.
//
//  Note that we have to handle saving on our own here.
//
function paste(doc, node) {
  if (doc.last_removed_node != null) {
    var new_node = doc.last_removed_node.cloneNode(true);
    new_node.style.display = "";
    node.parentNode.insertBefore(new_node, node);

    if (recording_a_script)
      script_commands[script_count++] =
	"platypus_do_function(window, 'script_paste',"+find_path(node)+
	","+find_path(doc.last_removed_node)+",null,null);";
  };
};

//
//  Script version of paste...
//
function script_paste(doc, where, what) {
    var new_node = what.cloneNode(true);
    new_node.style.display = "";
    where.parentNode.insertBefore(new_node, where);
};

//
//  Tue May 03 13:29:33 2005 -- Scott R. Turner
//
//  The interesting thing about Javascript is seeing how much you can
//  achieve without actually writing any code.  In this case, I'm able
//  to view the source for a page element by dint of the ever-handy
//  innerHTML and the data:uri trick.
//
//  Of course, this gives you the HTML in one ugly long string, so I'm
//  probably going to ruin this cleverness by introducing some
//  prettifying function at some point.  Damn me and my obsessive
//  attention to detail!
//
//  Wed May 04 15:30:53 2005 -- Scott R. Turner
//
//  I just came back to check and so far there's no prettifying going
//  on.  I seem to be overcoming my obsessive nature.
//
function view_source(doc, node) {
  doc.platypus_window.open('data:text/plain,'+
			   node.innerHTML,
			   'Platypus Source Viewer','');
};

//
//  isolate
//  Tue May 03 13:44:56 2005 -- Scott R. Turner
//
//  This is an interesting function: you basically want to erase
//  everything else on the page.  Now you might think that's easy to
//  do: just use "erase" on everything else in the DOM tree.
//  Unfortunately, erasing anything above this erases it as well.  So
//  that won't work.
//
//  A second thought is to pull this node up to the <body> level and
//  then remove or erase everything else.  However, I suspect for an
//  element like <TR> (a row of a table) that won't work.  But maybe,
//  so let's try it.
//
//  Hey, it works!  We can thank legions of bad Web page designers who
//  forced the Mozilla folks to correctly render bad HTML like rows
//  without tables.
//
function isolate(doc, node) {
  //
  //  Better not try this if we don't have a parent node.
  //
  if (!node.parentNode) return;
  //
  //  Remove node from its parent.
  //
  node.parentNode.removeChild(node);
  //
  //  Now remove everything else in the body.
  //
  while (doc.body.childNodes.length > 0) {
    doc.body.removeChild(doc.body.childNodes[0]);
  };
  //
  //  Now stick the original node back in.
  //
  doc.body.appendChild(node);
};

//
//  help
//  Tue May 03 14:38:23 2005 -- Scott R. Turner
//
//  My natural inclination is, of course, to leave the hapless user
//  with no clue at all how to use this extension.  But that would be
//  too cruel, even for me.  So as a compromise I'll include the help
//  as one long, obsfucated and unreadable string of HTML.  At least
//  I'll have the satisfaction of making your life hell.
//
//  Wed May 04 15:32:50 2005 -- Scott R. Turner
//
//  Damn, I just had to change the help file and it turns out I'm
//  really making my own life hell.  I hate that.
//
function help(doc, node) {
  doc.platypus_window.open('chrome://platypus/content/help.htm',
			   'Platypus Help','');
};

//
//  relax
//  Tue May 03 12:54:48 2005 -- Scott R. Turner
//
//  Remove all width restrictions from this element and any elements
//  underneath it.
//
function relax(doc, node) {
  walk_down(node, function (node) { node.style.width = 'auto';
                                    if (node.width) node.width = null; });
}

//
//  make_bw
//  Tue May 03 12:57:21 2005 -- Scott R. Turner
//
//  Turn this node and all underneath it B&W.
//
function make_bw(doc, node) {
  walk_down(node,
            function (node) { node.style.backgroundColor = "white";
			      node.style.color = "black";
			      node.style.backgroundImage = "";});
}

//
//  Scripting Commands
//
//  This is what got me into this whole mess in the first place.  I
//  had the bright idea of a user interface that would let you
//  manipulate a web page until it looked like you wanted, and then
//  would somehow save those manipulations until the next time you
//  loaded the page.  And then GreaseMonkey came along, and next thing
//  you know I'm learning Javascript and Firefox extensions.
//   
//  But enough about me, let's have more global variables!
//
//

//
//  start_scripting
//  Mon May 02 21:27:16 2005 -- Scott R. Turner
//
//  This is actually very simple.  We simply erase any old script
//  that's lying around, and flip the global flag that starts our
//  recording.
//
//  Note that this works perfectly fine to "reset" your recording.
//
function start_scripting(doc, node) {
    doc.script_commands = new Array;
    doc.script_count = 0;
    doc.recording_a_script = true;
};

//
//  save_script
//  Mon May 02 21:29:23 2005 -- Scott R. Turner
//
//  This is a little more complicated, but still pretty simple.  We
//  output some GreaseMonkey headers, and then walk through the
//  commands we've recorded, figuring out how to find the DOM element,
//  and then simply calling the same function on the element.  This is
//  possible because we require this extension to be loaded when the
//  recorded scripts are run (it doesn't work otherwise) and we export
//  to the GM scripts a function to access our functions.
//
//  We also have to be careful that our functions will run both as
//  part of this extension and when called from a GM script.
//
//  One other trick we use (and thanks to the GM guys for teaching me
//  this) is to load up our whole script in a data: URI.
//
function save_script(doc, node) {
    var i = 0;
    if (doc.script_count < 1) {
	alert("It doesn't appear to me that you've actually recorded anything, and I'm rarely wrong about things like that.");
	alert("Okay, I'm *occasionally* wrong.");
	return;
    };
    //
    //  Make sure he actually did something...
    //
    if (doc.script_count > 0) {
        var script = "";
	//
	//  Output the GM header.  Note that by default the new script
	//  will apply only to the current page.  User can edit that later
	//  via "Manage Scripts".
	//
	//  Also note that throughout this section as we build up the GM
	//  script we will insert newlines via %0d.  This is necessary because
	//  the whole script is going to end up being a huge data:uri, and
	//  there are restrctions on characters in URIs.
	//
        script += "// ==UserScript==";
        script += "%0d// @name            Platypus-"+doc.location;
        script += "%0d// @namespace       Platypus";
        script += "%0d// @include         "+doc.location;
	script += "%0dfunction do_platypus_script() {";
	while (i != doc.script_count) {
	    //
	    //  We've somewhat simplified this; the whole Javascript
	    //  statement is assembled by save_command, so all we need
	    //  do here is spit it out.
	    script += "%0d"+doc.script_commands[i++];
	};
    };
    script += "%0d}; // Ends do_platypus_script";
    //
    //  We have to hang the whole script on the load event because
    //  GM tends to break Firefox if you're too liberal in changing
    //  the DOM before the load has finished.
    //
    script += "%0dwindow.addEventListener(\"load\", function() { do_platypus_script() }, false);";

    //
    //  Here's the trick where we open the entire script as a data:uri.
    //  Note the crucial "//.user.js" at the end -- this looks like a 
    //  comment to Javascript and like a file extension to GM.  The result
    //  is that GM is tricked into accepting this as a script.
    //
    window.open('data:text/plain,'+script+'//.user.js','output','');

    //
    //  And now we erase our tracks, in case the FBI is misusing the
    //  Patriot Act.
    //
    stop_platypus(doc, node);
};

//
//  save_command
//  Mon May 02 21:43:00 2005 -- Scott R. Turner
//
//  Save a command for later scripting, if we're recording.  All
//  we save is the key and a path to the object.
//
//  You'll notice how unlike most Javascript authors, who would name
//  this function saveCommand, I use the name save_command.  In truth,
//  I disdain the sort of mixed capitalization style naming scheme
//  spawned by Java and its imitators.  It's clearly the work of
//  godless anarchists hell-bent on undermining the very foundations
//  of our culture -- the most important of those foundations being
//  good old Kernighan & Ritchie style C.
//
function save_command(doc, function_name, o, other) {
  if (doc.recording_a_script)
    doc.script_commands[doc.script_count++] =
      "platypus_do_function(window, '"+function_name+"',"+find_path(o)+",'"+other+"');";
};

//
//  platypus_do_function
//  Fri May 06 08:21:44 2005 -- Scott R. Turner
//
//  This is the actual function we export out of this package so that
//  the GM scripts we create can call Platypus functions.
//
//  We could do this a little easier with 'eval' but it would open
//  a potential security hole.
//
//  Mon May 09 11:26:23 2005 -- Scott R. Turner
//
//  The browser just threw me a curve-ball: "window" in an extension
//  gets reset whenever the user changes windows or tabs.  So when a
//  script makes a call back into platypus_do_function, "window" isn't
//  necessarily the same window the GM script was launched from.  That
//  makes perfect sense but kind of sucks from my point of view.  So
//  we need to make the GM script pass the window in to all the
//  scripted functions.
//
function platypus_do_function(win, func_name, o, other, other2, other3) {
  var func = null;

  //
  //  See if this function is in our Function table.
  //
  for (var i=0; i < FunctionInfo.length; i++)
    if (FunctionInfo[i][3] &&
	func_name == FunctionInfo[i][3].name) {
      func = FunctionInfo[i][3];
      break;
    } else if (FunctionInfo[i][5] &&
	       func_name == FunctionInfo[i][5].name) {
      func = FunctionInfo[i][5];
      break;
    };
  
  //
  //  No?
  //
  if (func == null) return;

  //
  //  Perhaps there's no node to operate upon?  While some functions
  //  might work without one, it's probably best not to go forward at
  //  this point.
  //
  if (!o) {
    Warning("Platypus couldn't find a page element when executing the command "+
	 func_name+".  This usually happens when running a script -- maybe the"+
	 " web page the script is running on has changed.");
    return;
  };

  //
  //  Sometimes win._content has no value?
  //
  if (!win._content) {
    Warning("Platypus had an error in executing the command "+func_name+
	 ".  Platypus was unable to access the win._content object.  This is "+
	 "an unusual error and should be reported to the Platypus maintainer at "+
	 "platypus.mozdev.org.  Thanks for your help!");
    return;
  };

  //
  //  And now apply the function
  //
  func(win._content.document, o, other, other2, other3);
  
};

//
//  FindPath
//  Mon May 02 21:47:26 2005 -- Scott R. Turner
//
//  You'll notice how here I've sneakily switched to the Java-style
//  function naming scheme.  This throws off the godless anarchists.
//  And possibly the FBI, if they're abusing the Patriot Act to spy
//  upon me.
//
//  Create an expression that will find an element in the page.  We
//  prefer to find an element by Id if that's possible (we assume that
//  ids are less likely to change than the page structure), otherwise
//  we construct an Xpath.  (We have to do this at the time the user edits
//  the element because some edits change the page structure.)
//
function find_path(o) {
    if (o.id != '') {
        return "document.getElementById('"+o.id+"')";
    } else {
	return "document.evaluate('"+
	    mybuildXPathForElement(o)+
	    "', document, null, XPathResult.FIRST_ORDERED_NODE_TYPE,"+
	    "null).singleNodeValue";
    };
};


//
//  Thu Apr 28 09:34:40 2005 -- Scott R. Turner
//
//  This function takes a page element and builds an Xpath to the
//  element by walking upward through the DOM tree appending
//  each element it finds.  We use this to identify the element
//  an editing function is applied to.
//
//  You know, this function puts in mind of the old Tom Lehrer song
//  "Lobachevsky":
//
//     I am never forget the day my first book is published.
//     Every chapter I stole from somewhere else.
//
//   I reminded of this because I stole this particular function from
//   some nameless Java programmer.  And by nameless I mean someone I
//   don't care to credit.  I do remember that I had to do some
//   rewriting, partly because the nameless author was so rude as to
//   write this in Java, and partly because XPath chokes if the
//   generated path goes all the way out to the root node.
//
//   Okay, I'll give him his props:
//
//   http://forum.java.sun.com/thread.jspa?threadID=612217&tstart=75
//
//   I don't know anything about this guy, other than he calls himself
//   "dmbdmb" and he's made 2,567 posts to a Java forum in the past
//   five years, which tends to suggest that he's a geek of major
//   proportions.  I'll give you odds he owns all the Star Trek movies.
//
//   Anyway, you've got to admire my honesty.  I could have taken full
//   credit for this and you never would have known.
//
//   In honor of dmbdmb, I've retained the asinine Java-style function
//   name.   Or should I say, InHonorOfDmbdmbI'veRetainedTheAsinine
//   JavaStyleFunctionName?
//
//   Thu May 12 10:24:32 2005 -- Scott R. Turner
//
//   Okay, here comes a subtlety.  XPaths work for documents, and
//   there should be one HTML document per web page.  But some web
//   sites apparently break this rule (www.nytimes.com for one).  Now
//   I'm not exactly sure what's going on here (but when am I ever?)
//   but the fix for now seems to be to stop at the HTML level and use
//   HTML[1] in all cases.
//
function mybuildXPathForElement(e) {
    var parent = e.parentNode;
    var path = "";

    if (e.nodeName == "HTML") {
      path += "/" + e.nodeName + "[1]";
    } else if (parent) {
	path = mybuildXPathForElement(parent);
	path += "/" + e.nodeName;

	var index = 1;
	var n = e.previousSibling;
	while (n != null) {
	    if (n.nodeName == e.nodeName) index++;
	    n = n.previousSibling;
	}
	path += "[" + index + "]";
    }
    
    return path;
}

//
//  center
//  Wed May 04 16:26:10 2005 -- Scott R. Turner
//
//  Wrap the selected element in a <CENTER>
//
function center(doc, node) {
  // 
  //  Create a "center" node.
  //
  var center_node = doc.createElement ("CENTER");
  //
  //  Now stick him in right before (or after, doesn't matter) of
  //  the current element.
  //
  //  (This comment stolen verbatim from above.)
  //
  node.parentNode.insertBefore(center_node, node);
  //
  //  Now remove node from its current parent and stick it into
  //  the center node.
  //
  node.parentNode.removeChild(node);
  center_node.appendChild(node);
  //
  //  Put the focus on the center node.
  //
  focus_on(doc, center_node);
};

//
//  Thu May 05 14:36:45 2005 -- Scott R. Turner
//
//  The style-setting functions.  The idea here is to pop up a dialog,
//  let the user select the styles he wants for the element, and then
//  apply them.
//
//  There are a couple of tricky things (beyond the whole trickiness
//  of popping up a dialog box and doing a callback into this
//  Javascript, which I'm skipping over in order to impress you with
//  my casual expertise) about this.
//
//  First, we have to be careful not to conflict with the style-munging
//  that is going on as part of the highlighting.
//
//  Second, we need to handle the scripting for this a little
//  differently, because we're not just saving the name of a function
//  and the referenced object; we also need the changed style
//  attributes. 
//
//
function set_style(doc, node) {
    //
    //  Before we actually edit the styles of the element, save
    //  the existing styles so that we can reset if necessary.
    //
    doc.saved_style = doc.focus.getAttribute('style');
    //
    //  I can't actually find openDialog documented anywhere, so
    //  I'm not sure exactly how it works.
    //
    doc.platypus_window.openDialog("chrome://platypus/content/style-editor.xul", 
	 	      "Style Editor", "resizable,centerscreen,modal");
};

//
//  This is hooked up to the "Apply" button on the style editor
//  dialog.  When you hit "Apply" the changes happen on-screen.
//
function apply_style(doc, new_style) {
    doc.focus.setAttribute('style', new_style);
};

//
//  This is like apply, but we save the command.
//
function accept_style(doc, new_style) {
    doc.focus.setAttribute('style', new_style);
    save_command(doc, 'set_style_script', doc.focus, new_style);
};

//
//  Script version of function.
//
function set_style_script(doc, element, new_style) {
    element.setAttribute('style', new_style);
};

//
//  cancel_style gets called if (do you begin to sense a pattern?) hit
//  the "Cancel" button on the style editor.  This is the case where
//  we need to restore the original style.
//
function cancel_style(doc) {
    doc.focus.setAttribute('style',doc.saved_style);
};

//
//  Modify URL
//
//
function modify_url(doc, node) {
  if (node.nodeName != 'A' ||
      node.href == null) {
    alert("You need to use this function on a link, and you don't "+
	  "currently have a link selected.  You'll know you have a "+
	  "link selected when you see an 'A' or the link location "+
	  "appear in the "+
	  "status bar at the bottom of the browser.  If you're having "+
	  "trouble selecting the URL, try using the arrow keys to "+
	  "find the URL.");
  } else {
    window.openDialog("chrome://platypus/content/url-regexp.xul", 
	 	      "Modify URL", "resizable,centerscreen,modal");
  };
};

//
//  This function is a callback from the Modify URL popup.  It applies
//  the regexp to either the focus URL or all the URLs on the page.
//
function do_modify_url(doc, node, match_re, replace_string, global_flag) {
    Dump("In do_url, match_re = "+match_re+" replace_string = "+replace_string+
	 " global flag = "+global_flag);
    //
    //  Need to hand the match regexp actually coming in as
    //  a string.  This should handle it regardless!
    //
    match_re = new RegExp(match_re);

    if (global_flag) {
	var allurls =
	    doc.getElementsByTagName('A');
	Dump("  ...modifying all URLS: "+allurls.length);
	for(var i = 0; i < allurls.length; i++)
	    modify_single_url(doc, match_re, replace_string, allurls[i]);
    } else {
	modify_single_url(doc, match_re, replace_string, node);
    };
    //
    //  We implement our own save_command here, since this doesn't fit
    //  in with the generic one.
    //
    if (doc.recording_a_script)
	doc.script_commands[doc.script_count++] =
	  "platypus_do_function(window, 'do_modify_url',"+find_path(doc.focus)+
	  ","+match_re+",'"+replace_string+"',"+global_flag+");";
};

function modify_single_url(doc, match_re, replace_string, node) {
    if (node.href) {
	node.href = node.href.replace(match_re, replace_string);
    };
};

//
//  Modify HTML
//
//
function modify_html(doc, node) {
    window.openDialog("chrome://platypus/content/html-regexp.xul", 
	 	      "Modify HTML", "resizable,centerscreen,modal");
};

//
//  This function is a callback from the Modify HTML popup.  It applies
//  the regexp to the focus HTML.
//
function do_modify_html(doc, element, match_re, replace_string) {
    match_re = new RegExp(match_re);
    if (element.innerHTML) {
	element.innerHTML = element.innerHTML.replace(match_re, replace_string);
    if (doc.recording_a_script)
	doc.script_commands[doc.script_count++] =
	  "platypus_do_function(window, 'do_modify_html',"+find_path(doc.focus)+
	  ","+match_re+",'"+replace_string+"',null);";
    };
};

//
//  Insert HTML
//
//
function insert_html(doc, node) {
    window.openDialog("chrome://platypus/content/arbitrary-insert.xul", 
	 	      "Modify HTML", "resizable,centerscreen,modal");
};

//
//  Stupid DOM designers.
//
function insertAfter(newNode, target) {
    var parent = target.parentNode;
    var refChild = target.nextSibling;

    Dump("In insertAfter, parent = "+parent+" refChild = "+refChild);
    
    if(refChild != null)
	parent.insertBefore(newNode, refChild);
    else
	parent.appendChild(newNode);
};

//
//  This function is a callback from the Modify HTML popup.  It applies
//  the regexp to the focus HTML.
//
function do_insert_html(doc, element, new_html, before, insert_as_block) {
  var new_element;

  Dump("In do_insert_html, element = "+element+" new_html = "+new_html);
  Dump("   before = "+before+" insert_as_block = "+insert_as_block);
  if (insert_as_block) {
    new_element = doc.createElement ("DIV");
  } else {
    new_element = doc.createElement ("SPAN");
  };
  new_element.innerHTML = new_html;

  Dump("New element = "+new_element);
  
  if (before) {
      element.parentNode.insertBefore(new_element, element);
  } else {
      insertAfter(new_element, element);
  };
  if (doc.recording_a_script)
    doc.script_commands[doc.script_count++] =
      "platypus_do_function(window, 'do_insert_html',"+find_path(element)+
      ",'"+new_html+"',"+before+","+insert_as_block+");";
};

//
//  Fri May 06 12:30:13 2005 -- Scott R. Turner
//
//  There's no undo...
//
function no_undo() {
  alert("I'm sorry, there's no Undo function in Platypus."+
	"  Reloading the page will undo all the unsaved changes.");
};

//
//  Function Dispatch Table
//
//  This array serves as a function dispatch table.  We can look up
//  a keypress here and find the associated function.  This probably
//  should be an array of structures (objects) but I'm lazy and don't
//  really want to learn enough Javascript to create static objects on
//  fly to initialize an array. 
//  
//  Key, Function, Save
//
//  Key -- Keypress to initiate function.
//  CtlKey -- true if ctl+key
//  AltKey -- true if alt+key
//  Function -- Reference to the function.
//  Save -- If true, save this command when scripting.
//  Saved_Function -- If a different function is called when this is
//                    run from a GM script, this is it...
//
var FunctionInfo =
  [
   ['Q', false, false, stop_platypus, false, null],
   ['q', false, false, stop_platypus, false, null],
   ['B', false, false, make_bw, true, null],
   ['b', false, false, make_bw, true, null],
   ['C', false, false, center, true, null],
   ['c', false, false, center, true, null],
   ['E', false, false, erase, true, null],       
   ['e', false, false, erase, true, null],       

   // Delete key & Ctl-X for Cut
   [46,  false, false, remove, true, null],	       
   ['x', true, false, remove, true, null],	       
   ['X', true, false, remove, true, null],

   // Ctl-V for paste.
   ['V', true, false, paste, false, script_paste],
   ['v', true, false, paste, false, script_paste],

   ['V', false, false, view_source, false, null],
   ['v', false, false, view_source, false, null],

   ['I', false, false, isolate, true, null],
   ['i', false, false, isolate, true, null],
   
   // Ctl-S to save (a script)
   ['S', true, false, save_script, false, null],
   ['s', true, false, save_script, false, null],
   
   // Help
   ['?', false, false, help, false, null],
   
   // Up Arrow
   [38,  false, false, navigate_wider, false, null],
   ['<', false, false, navigate_wider, false, null],
   // Down Arrow
   [40,  false, false, navigate_narrower, false, null],
   ['>', false, false, navigate_narrower, false, null],
   // Right Arrow
   [39,  false, false, navigate_next_sibling, false, null],
   ['+', false, false, navigate_next_sibling, false, null],
   // Left Arrow
   [37,  false, false, navigate_previous_sibling, false, null],
   ['-', false, false, navigate_previous_sibling, false, null],

   ['!', false, false, set_style, false, set_style_script],
   ['S', false, false, set_style, false, set_style_script],
   ['s', false, false, set_style, false, set_style_script],
   
   ['@', false, false, modify_url, false, do_modify_url],
   ['U', false, false, modify_url, false, do_modify_url],
   ['u', false, false, modify_url, false, do_modify_url],

   // Modify HTML
   ['#', false, false, modify_html, false, do_modify_html],
   ['M', false, false, modify_html, false, do_modify_html],
   ['m', false, false, modify_html, false, do_modify_html],

   ['$', false, false, insert_html, false, do_insert_html],
   ['H', false, false, insert_html, false, do_insert_html],
   ['h', false, false, insert_html, false, do_insert_html],

   ['z', true, false, no_undo, false, null],
   ['Z', true, false, no_undo, false, null],

   ['R', false, false, relax, true, null],
   ['r', false, false, relax, true, null]];


//
//  Generic functions to work with FunctionInfo
//
function find_function(key) {
  for (var i=0; i < FunctionInfo.length; i++)
    if (key == FunctionInfo[i][0])
      return FunctionInfo[i][3];
  return null;
};

function handle_keypress(evt) {
  var c;
  var ctrl, alt;
  var entry = null;
  var doc =  evt.currentTarget;

  Dump(" In evt doc = "+doc);
  if (evt.keyCode != 0) c = evt.keyCode;
  else c = String.fromCharCode(evt.which);
  ctrl = evt.ctrlKey;
  alt = evt.altKey;
  Dump("Keypress: "+c+" Ctrl: "+ctrl+" Alt: "+alt);
  //
  //
  //  Now dispatch on the keypress char.  We repeat some code that
  //  we've already written in find_function above.  It's a little
  //  sub-optimal, but this way we can hang on to the index into the
  //  function table, which we also need to see if have to save the
  //  command.
  //
  for (var i=0; i < FunctionInfo.length; i++)
      if (c == FunctionInfo[i][0]
	  && (FunctionInfo[i][1] == ctrl)
	  && (FunctionInfo[i][2] == alt)) {
	  entry = i;
	  Dump("Found entry for key: "+FunctionInfo[i][3].name);
	  break;
      };
  //
  //  Dang, dang, dang!  I just realized that all these user commands
  //  that I've written to work on "focus" need to be re-written to
  //  work on an argument so that they can be re-used in the scripts.
  //  How stupid is that?  I really shouldn't be allowed anywhere near
  //  a keyboard.  I'm a danger to myself and compilers everywhere.
  //
  //  Well, excuse me for a minute while I go rewrite all those
  //  functions.
  //
  //  Okay, I'm back.  Now where was I?  Oh yes.
  //
  if (entry != null) {
    //
    //  We need this global because a few of our user functions
    //  need to reach back and figure out what key called them.
    //
    doc.last_keypress = c;
    try {
      if (FunctionInfo[entry][4])
	if (FunctionInfo[entry][5]) {
	  save_command(doc, FunctionInfo[i][5].name,doc.focus,null,null,null);
	} else {
	  save_command(doc,FunctionInfo[i][3].name,doc.focus,null,null,null);
	};
      FunctionInfo[i][3](doc, doc.focus);
      //
      //  Since we handled this keypress, stop this event dead in
      //  its tracks so that no other handler ever sees it.  This
      //  turns out to be somewhat hard to do...
      //
      evt.preventDefault();
      evt.stopPropagation();
      evt.cancelBubble = true;
      return false;
    } catch (e) {};
  };
};

//
//  handle_mouseover
//  Sat Apr 30 23:49:21 2005 -- Scott Turner
//
//  When we mouseover something, we have to switch our focus to the
//  new object.
//
function handle_mouseover(evt) {
  Dump("In handle_mouseover with evt = "+evt);
  var new_elem = (evt.target) ? evt.target : evt.srcElement;
  Dump("  New_elem = "+new_elem);
  Dump("  ownerDocument = "+new_elem.ownerDocument);
  focus_on(new_elem.ownerDocument, new_elem);
};

//
//  startPlatypus
//  Sat Apr 30 23:18:42 2005 -- Scott R. Turner
//
//  When Platypus starts up, it hangs event handlers on keypress (to
//  handle user inputs), on mouseover events (to handle the user
//  pointing to some new part of the page) and puts the Platypus
//  commands on the context menu as well.
//
function start_platypus()
{
  Dump("Platypus_started = "+window._content.document.title);
  Dump("Platypus_started = "+window._content.document.location);
  if (window._content.document.platypus_started != true) {
    Dump("In start_platypus.");
    window._content.document.addEventListener("keypress",
					      handle_keypress, false);
    Dump("After adding keypress.");

    window._content.document.addEventListener("mouseover",
					      handle_mouseover, false);
    Dump("After adding mouseover.");

    window._content.document.addEventListener("onunload",
					      stop_platypus, false);
    Dump("After adding onunload.");
    window.captureEvents(Event.KeyPress);
    Dump("After captureEvents.");
    //
    //  Initialize all the "global" variables for this window.
    //
    window._content.document.interactive = true;
    window._content.document.platypus_started = true;
    window._content.document.focus = null;
    window._content.document.last_removed_node = null;
    window._content.document.script_commands = null;
    window._content.document.script_count = 0;
    window._content.document.recording_a_script = false;
    window._content.document.saved_style = null;
    window._content.document.last_keypress = null;
    window._content.document.platypus_window = window;
    start_scripting(window._content.document, null);
    Dump("End of Platypus initialization.");
  };
}    

//
//  stopPlatypus
//  Sat Apr 30 23:18:42 2005 -- Scott R. Turner
//
//  Remove the event handles and context menu items.
//
function stop_platypus(doc, node)
{
  Dump("Stop platypus: "+doc.title+", "+doc.location+", "+doc.platypus_window);
  if (doc.platypus_started == true) {
    Dump("Clearing focus, etc.");
    clear_focus(doc);
    doc.removeEventListener("keypress", handle_keypress, false);
    doc.removeEventListener("mouseover", handle_mouseover, false);
    doc.platypus_window.releaseEvents(Event.KeyPress);
    doc.interactive = false;
    doc.platypus_started = false;
  };
}    

